from sqlalchemy.engine import create_engine
from sqlalchemy.exc import ProgrammingError, DBAPIError
from sqlalchemy.orm import sessionmaker, mapper, relation, backref
from sqlalchemy.sql import text, expression, bindparam
from sqlalchemy import Table, Column, Integer, String, MetaData, ForeignKey

engine = create_engine("virtuoso://dba:dba@VOS")
Session = sessionmaker(bind=engine)
session = Session(autocommit=False)
metadata = MetaData()

test_table = Table('test_table', metadata,
                   Column('id', Integer, primary_key=True),
                   Column('name', String),
                   schema="test"
                   )
class Object(object):
    def __init__(self, **kw):
        for k,v in kw.items():
            setattr(self, k, v)

test_table_a = Table("test_a", metadata,
                     Column("id", Integer, primary_key=True),
                     Column('name', String),
                     schema="test")
test_table_b = Table("test_b", metadata,
                     Column("id", Integer, primary_key=True),
                     Column('name', String),
                     Column("a_id", Integer, ForeignKey(test_table_a.c.id)),
                     schema="test")
class A(Object): pass
class B(Object): pass

def table_exists(table):
    conn = engine.connect()
    result = conn.execute(
        text("SELECT TABLE_NAME FROM TABLES WHERE "
             "lower(TABLE_SCHEMA) = :schemaname AND "
             "lower(TABLE_NAME) = :tablename",
             bindparams=[
                 bindparam("tablename", table.name),
                 bindparam("schemaname", table.schema if table.schema else "DBA")
                 ])
        )
    return result.scalar() is not None

def clean():
    for table in ("test_table", "test_b", "test_a"):
        conn = engine.connect()
        result = conn.execute(
            text("SELECT 1 FROM TABLES WHERE "
                 "lower(TABLE_NAME) = '%s'" % table)
            )
        if result.scalar():
            conn.execute(text("DROP TABLE %s" % table))

class Test01Basic(object):
    @classmethod
    def setup_class(self):
        clean()

    def test_01_table(self):
        test_table.create(engine)
        try:
            assert table_exists(test_table)
        finally:
            test_table.drop(engine)
        assert not table_exists(test_table)

    def test_02_fkey(self):
        test_table_a.create(engine)
        test_table_b.create(engine)
        try:
            test_table_a.drop(engine)
            assert False, "Should not be able to drop %s because of FKEY" % test_table_a
        except DBAPIError:
            pass
        test_table_b.drop(engine)
        test_table_a.drop(engine)

class Test02Object(object):
    @classmethod
    def setup_class(cls):
        clean()
        test_table.create(engine)
        mapper(Object, test_table)

    def test_01_insert(self):
        o1 = Object(name = "foo")
        session.add(o1)
        o2 = Object(name = "bar")
        session.add(o2)
        session.commit()

        o = session.query(Object).get(1)
        assert o.name == "foo"
        o = session.query(Object).get(2)
        assert o.name == "bar"

    def test_02_update(self):
        assert session.query(Object).filter(Object.name == "foo").count() == 1

        o = session.query(Object).filter(Object.name == "foo").all()[0]
        _oid = o.id
        o.name = "baz"
        session.add(o)
        session.commit()

        assert session.query(Object).filter(Object.name == "foo").count() == 0
        assert session.query(Object).filter(Object.name == "baz").count() == 1

        o = session.query(Object).filter(Object.name == "baz").all()[0]
        assert o.id == _oid
        
    def test_03_delete(self):
        [session.delete(o) for o in
         session.query(Object).filter(Object.name == "baz").all()]
        session.commit()

        assert session.query(Object).filter(Object.name == "baz").count() == 0
        assert session.query(Object).count() > 0

        [session.delete(o) for o in
         session.query(Object).all()]
        session.commit()

        assert session.query(Object).count() == 0

class Test03Relation(object):
    @classmethod
    def setup_class(cls):
        clean()
        test_table_a.create(engine)
        test_table_b.create(engine)
        mapper(A, test_table_a)
        mapper(B, test_table_b, properties={ 'a': relation(A) })

    def test_01_create(self):
        a = A()
        b = B()
        b.a = a
        # NB, a gets implicitly added
        session.add(b)
        session.commit()
    
        b = session.query(B).get(1)
        assert isinstance(b.a, A)

    def test_02_update(self):
        c = A()
        b = session.query(B).get(1)
        _oldid = b.a.id
        b.a = c
        session.add(b)
        session.commit()

        b = session.query(B).get(1)
        assert isinstance(b.a, A)
        assert b.a.id != _oldid

    def test_03_fkey_violation(self):
        b = session.query(B).get(1)
        # delete out from under, should raise a foreign key
        # constraint because we haven't set cascade on the
        # relation
        try:
            session.delete(b.a)
            session.commit()
            raise ValueError("Should have had an exception because of FK")
        except:
            session.rollback()

    def test_04_delete(self):
        b = session.query(B).get(1)
        # delete out from under, should raise a foreign key
        # constraint because we haven't set cascade on the
        # relation
        session.delete(b)
        session.delete(b.a)
        session.commit()

        assert session.query(A).count() == 1
        assert session.query(B).count() == 0



from rdflib.graph import ConjunctiveGraph, Graph
from rdflib.store import Store
from rdflib.plugin import get as plugin
from rdflib.namespace import RDF, RDFS
from rdflib.term import URIRef, Literal, BNode
from datetime import datetime
from virtuoso.vstore import Virtuoso

class Test00Plugin(object):
    def test_get_plugin(self):
        V = plugin("Virtuoso", Store)
        assert V is Virtuoso

from math import pi
test_statements = (
    (URIRef("http://example.org/"), RDF["type"], RDFS["Resource"]),
    (BNode(), RDF["type"], RDFS["Resource"]),
    (URIRef("http://example.org/"), RDF["type"], BNode()),
    (URIRef("http://example.org/"), RDFS["label"], Literal("hello world")),
    (URIRef("http://example.org/"), RDFS["comment"],
     Literal("Here we have a long comment to purposely overflow the inline RDF_QUAD limit. "
             "We keep talking and talking, but what are we saying? Precisely nothing the "
             "whole idea is to have a bunch of characters here. Blah blah, yadda yadda, "
             "etc. This is probably enough. Hopefully. One more sentence to make certain.")),
    (URIRef("http://example.org/"), RDFS["label"], Literal(3)), # Fails because comes back untyped
    (URIRef("http://example.org/"), RDFS["comment"], Literal(datetime.now())),
    (URIRef("http://example.org/"), RDFS["comment"], Literal(datetime.now().date())),
#    (URIRef("http://example.org/"), RDFS["label"], Literal(pi)), # Fails because floats cannot be found?
    (URIRef("http://example.org/"), RDFS["label"], Literal("hello world", lang="en")), # Fails because comes back w/o language
    )

class Test01Store(object):
    @classmethod
    def setup_class(cls):
        cls.store = Virtuoso("DSN=VOS;UID=dba;PWD=dba;WideAsUTF16=Y")
        cls.graph = Graph(cls.store, identifier=URIRef("http://example.org/"))
            
    @classmethod
    def teardown_class(cls):
        cls.store.sparql_query("CLEAR GRAPH %s" % cls.graph.identifier.n3())
        cls.store.commit()
        cls.store.close()
        
    def test_01_query(self):
        g = ConjunctiveGraph(self.store)
        count = 0
        for statement in g.triples((None, None, None)):
            count += 1
            break
        assert count == 1, "Should have found at least one triple"

    def test_02_contexts(self):
        g = ConjunctiveGraph(self.store)
        for c in g.contexts():
            assert isinstance(c, Graph)
            
    def add_remove(self, statement):
        # add and check presence
        self.graph.add(statement)
        self.store.commit()
        
        assert statement in self.graph, "%s not found" % (statement,)

        # check that we really got back what we asked for
        for x in self.graph.triples(statement):
            assert statement == x, "Round-trip mismatch:\n\t%s\n\t%s" % (statement, x)

        # delete and check absence
        self.graph.remove(statement)
        self.store.commit()
        
        assert statement not in self.graph, "%s found" % (statement,)

# make separate tests for each of the test statements so that we don't
# get flooded with unreadable and irrelevant log messages if one fails
def _mk_add_remove(name, s):
    def _f(self):
        self.add_remove(s)
    _f.func_name = name
    return _f
for i in range(len(test_statements)):
    attr = "test_%02d_add_remove" % (i + 10)
    setattr(Test01Store, attr, _mk_add_remove(attr, test_statements[i]))




from setuptools import setup, find_packages
import sys, os

version = '0.6'
try:
    from mercurial import ui, hg, error
    repo = hg.repository(ui.ui(), ".")
    ver = repo[version]
except ImportError:
    pass
except error.RepoLookupError:
    tip = repo["tip"]
    version = version + ".%s.%s" % (tip.rev(), tip.hex()[:12])
except error.RepoError:
    pass
                            
def readme():
    dirname = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(dirname, "README.txt")
    return open(filename).read()

setup(name='virtuoso',
      version=version,
      description="OpenLink Virtuoso Support for SQLAlchemy and RDFLib",
      long_description=readme(),
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='',
      author='Open Knowledge Foundation',
      author_email='okfn-help@okfn.org',
      url='http://packages.python.org/virtuoso',
      license='BSD',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          # -*- Extra requirements: -*-
          "pyodbc>=2.1.6",
          "SQLAlchemy>=0.6.4",
          "rdflib"
      ],
      entry_points="""
          [sqlalchemy.dialects]
          virtuoso = virtuoso:alchemy.VirtuosoDialect

          [rdf.plugins.store]
          Virtuoso = virtuoso:vstore.Virtuoso
      """,
      )



try:
    from rdflib.graph import Graph
    from rdflib.term import URIRef, BNode, Literal, Variable
except ImportError:
    from rdflib.Graph import Graph
    from rdflib import URIRef, BNode, Literal, Variable
from rdflib.store import Store, VALID_STORE, NO_STORE
from traceback import format_exc
import pyodbc

__all__ = ['Virtuoso', 'resolve']

log = __import__("logging").getLogger(__name__)

class Virtuoso(Store):
    context_aware = True
    transaction_aware = True
    def __init__(self, *av, **kw):
        super(Virtuoso, self).__init__(*av, **kw)
        self.__prefix = {}
        self.__namespace = {}

    def open(self, dsn):
        try:
            self._connection = pyodbc.connect(dsn)
            self._cursor = self._connection.cursor()
            log.info("Virtuoso Store Connected: %s" % dsn)
            return VALID_STORE
        except:
            log.error("Virtuoso Connection Failed:\n%s" % format_exc())
            return NO_STORE

    def close(self, commit_pending_transaction=False):
        if commit_pending_transaction:
            self.commit()
        else:
            self.rollback()
        self._connection.close()

    def query(self, q):
        try:
            return self._cursor.execute(q.encode('utf-8')) #str(q))
        except:
            log.error(u"Error Executing: %s" % q)
            raise

    def sparql_query(self, q):
        return self.query(u"SPARQL " + q)
    
    def commit(self):
        self.query("COMMIT WORK")
    def rollback(self):
        self.query("ROLLBACK WORK")

    def contexts(self, statement=None):
        if statement is None:
            q = (u'SELECT DISTINCT __ro2sq(G) FROM RDF_QUAD')
        else:
            q = (u'SELECT DISTINCT ?g WHERE '            
                 u'{ GRAPH ?g { %(S)s %(P)s %(O)s } }')
            q = _query_bindings(statement)
        for uri, in self.query(q):
            yield Graph(self, identifier=URIRef(uri))
            
    def triples(self, statement, context=None):
        query_bindings = _query_bindings(statement, context)
        for k,v in query_bindings.items():
            if v.startswith('?') or v.startswith('$'):
                query_bindings[k+"v"]=v
            else:
                query_bindings[k+"v"]="(%s) AS %s" % (v, Variable(k).n3())
        q = (u'SPARQL define output:valmode "LONG" '
             u'SELECT DISTINCT %(Sv)s %(Pv)s %(Ov)s %(Gv)s '
             u'WHERE { GRAPH %(G)s { %(S)s %(P)s %(O)s } }')
        q = q % query_bindings
        log.debug(q)

        resolver = self._connection.cursor()
        for result in self.query(q):
            s,p,o,g = [resolve(resolver, x) for x in result]
            yield (s,p,o), g
        resolver.close()

    def add(self, statement, context=None, quoted=False):
        assert not quoted, "No quoted graph support in Virtuoso store yet, sorry"
        query_bindings = _query_bindings(statement, context)
        q = u'SPARQL INSERT '
        if context is not None:
            q += u'INTO GRAPH %(G)s ' % query_bindings
        q += u'{ %(S)s %(P)s %(O)s }' % query_bindings
        log.debug(q)
        self.query(q)

    def remove(self, statement, context=None, quoted=False):
        assert not quoted, "No quoted graph support in Virtuoso store yet, sorry"
        if statement == (None, None, None) and context is not None:
            q = u'SPARQL CLEAR GRAPH %s' % context.identifier.n3()
        else:
            query_bindings = _query_bindings(statement, context)
            q = u'SPARQL DELETE '
            if context is not None:
                q += u'FROM GRAPH %(G)s ' % query_bindings
            q += u'{ %(S)s %(P)s %(O)s }' % query_bindings
            log.debug(q)
        self.query(q)

    def bind(self, prefix, namespace):
        self.__prefix[namespace] = prefix
        self.__namespace[prefix] = namespace

    def namespace(self, prefix):
        return self.__namespace.get(prefix, None)

    def prefix(self, namespace):
        return self.__prefix.get(namespace, None)

    def namespaces(self):
        for prefix, namespace in self.__namespace.iteritems():
            yield prefix, namespace

def _bnode_to_nodeid(bnode):
    from string import ascii_letters
    iri = bnode
    for c in bnode[1:]:
        if c in ascii_letters:
            # from rdflib not virtuoso
            iri = "b" + "".join(str(ord(x)-38) for x in bnode[:8])
            break
    return URIRef("nodeID://%s" % iri)

def _nodeid_to_bnode(iri):
    from string import digits
    iri = iri[9:] # strip off "nodeID://"
    bnode = iri
    if len(iri) == 17:
        # assume we made it...
        ones, tens = iri[1::2], iri[2::2]
        chars = [x+y for x,y in zip(ones, tens)]
        bnode = "".join(str(chr(int(x)+38)) for x in chars)
    return BNode(bnode)

def resolve(resolver, args):
    """
    Takes the Virtuoso representation of an RDF node and returns
    an appropriate instance of :class:`rdflib.term.Node`.

    :param resolver: an cursor that can be used for database
        queries necessary to resolve the value
    :param args: the tuple returned
        by :mod:`pyodbc` in case of a SPASQL query.
    """
    (value, dvtype, flag, lang, dtype) = args
#    if dvtype in (129, 211):
#        print "XXX", dtype, value, dtype
    if dvtype == pyodbc.VIRTUOSO_DV_IRI_ID:
        q = (u'SELECT __ro2sq(%s)' % value)
        resolver.execute(str(q))
        iri, = resolver.fetchone()
        if iri.startswith("nodeID://"):
            return _nodeid_to_bnode(iri)
        return URIRef(iri)
    if dvtype == pyodbc.VIRTUOSO_DV_RDF:
        return Literal(value, lang=lang or None, datatype=dtype or None)
    if dvtype == pyodbc.VIRTUOSO_DV_STRING:
        return Literal(value)
    if dvtype == pyodbc.VIRTUOSO_DV_LONG_INT:
        return Literal(int(value))
    if dvtype == pyodbc.VIRTUOSO_DV_SINGLE_FLOAT:
        return Literal(value, "http://www.w3.org/2001/XMLSchema#float")
    if dvtype == pyodbc.VIRTUOSO_DV_DATETIME:
        return Literal(value.replace(" ", "T"),
                       datatype=URIRef("http://www.w3.org/2001/XMLSchema#dateTime"))
    if dvtype == pyodbc.VIRTUOSO_DV_DATE:
        return Literal(value, datatype=URIRef("http://www.w3.org/2001/XMLSchema#date"))
    log.warn("Unhandled SPASQL DV type: %d for %s" % (dvtype, value))
    return Literal(value)

def _query_bindings((s,p,o), g=None):
    if isinstance(g, Graph):
        g = g.identifier
    if s is None: s = Variable("S")
    if p is None: p = Variable("P")
    if o is None: o = Variable("O")
    if g is None: g = Variable("G")
    if isinstance(s, BNode):
        s = _bnode_to_nodeid(s)
    if isinstance(p, BNode):
        p = _bnode_to_nodeid(p)
    if isinstance(o, BNode):
        o = _bnode_to_nodeid(o)
    if isinstance(g, BNode):
        g = _bnode_to_nodeid(g)
    return dict(
        zip("SPOG", [x.n3() for x in (s,p,o,g)])
        )



from sqlalchemy.connectors.pyodbc import PyODBCConnector
from sqlalchemy.dialects.sybase.base import SybaseDialect
from sqlalchemy.sql import text, expression, bindparam
from sqlalchemy import types as sqltypes
from sqlalchemy.engine import default, reflection

class VirtuosoExecutionContext(default.DefaultExecutionContext):
    def get_lastrowid(self):
        self.cursor.execute("SELECT identity_value() AS lastrowid")
        lastrowid = self.cursor.fetchone()[0]
        return lastrowid
    
class VirtuosoDialect(PyODBCConnector, SybaseDialect):
    execution_ctx_cls = VirtuosoExecutionContext
    
    def initialize(self, connection):
        self.supports_unicode_statements = False
        self.supports_unicode_binds = False
        SybaseDialect.initialize(self, connection)
        
    def _get_default_schema_name(self, connection):
        return 'DBA'

    def has_table(self, connection, tablename, schema=None):
        if schema is None:
            schema = self.default_schema_name
        result = connection.execute(
            text("SELECT TABLE_NAME FROM TABLES WHERE "
                 "TABLE_SCHEMA=:schemaname AND "
                 "TABLE_NAME=:tablename",
                 bindparams=[
                     bindparam("schemaname", schema),
                     bindparam("tablename", tablename)
                     ])
            )
        return result.scalar() is not None
    
    def get_table_names(self, connection, schema=None, **kw):
        if schema is None:
            schema = self.default_schema_name
        result = connection.execute(
            text("SELECT TABLE_NAME FROM TABLES WHERE TABLE_SCHEMA=:schemaname",
                 bindparams=[bindparam("schemaname", schema)])
            )
        return [r[0] for r in result]

